/* Mptn -- regexp-like pattern matcher suited for morphological analysis
 * Copyright (C) 2000  George Bronnikov
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
  Operations for particular expression types.

  $Id: mptn-ops.c 1.3 Sun, 23 Jan 2000 20:56:51 -0600 goga $
*/

#include <ctype.h>

#include "mptn.h"
#include "mptn-chrset.h"
#include "mptn-private.h"

static void make_offset (FILE *fp, guint offset);

/*************************** A constant string ********************************/

static void string_matcher_destroy (gpointer mctl);
static gpointer string_matcher_iter_start (mptn_t *mptn,
					   chr *begin, chr *end,
					   mptn_var_t *vars, guint stage,
					   gpointer data);
static chr *string_matcher_subst (mptn_t *mptn, mptn_var_t *vars, gpointer data);
static void string_matcher_print (mptn_t *mptn, FILE *fp, guint offset);

mptn_matcher_ops_t mptn_string_matcher_ops = {
  mptn_string_matcher_init,
  string_matcher_destroy,
  string_matcher_iter_start,
  NULL,				/* Since we can only have one match, 
				   no need for iter_step. */
  NULL,				/* Since we allocate no memory, 
				   no need for iter_free.*/
  NULL,				/* No stage changes here, 
				   thus no stage function. */
  string_matcher_subst,
  string_matcher_print
};

/*
  Initialization
*/
mptn_t *
mptn_string_matcher_init (mptn_control_t *control, 
			  chr *args_begin, chr *args_end,
			  GArray *varnames,
			  gpointer data)
{
  mptn_t *res;

  g_assert (args_begin != NULL);
  control = mptn_control_normalize (control);

  if (!args_end)
    args_end = args_begin + chra_len (args_begin);

  res = mptn_new (control);
  res->ops = &mptn_string_matcher_ops;

  
  /* 
     There is just one arg, and this is the string.
     We should just copy it (who knows in what kind of memory it
     is kept).
   */
  res->mctl =  chra_ndup (args_begin, args_end - args_begin);
  res->min_len = res->max_len = chra_len (res->mctl);

  return res;
}

/*
  Matcher destruction.
*/
static void
string_matcher_destroy (gpointer mctl)
{
  g_free (mctl);
}

/*
  Iterator start function.
  In this case, we just check the string against the pattern.
  In the case of success, return (gpointer)1;
  in the case of failure, NULL.
*/
static gpointer
string_matcher_iter_start (mptn_t *mptn,
			   chr *begin, chr *end,
			   mptn_var_t *vars, guint stage,
			   gpointer data)
{
  chr * ptn_string;

  MPTN_CHECK (mptn);
  g_assert (begin != NULL);
  g_assert (end != NULL);
  
  ptn_string = mptn->mctl;;

  if (mptn->min_len == end - begin
      && chra_ncmp (ptn_string, begin, end - begin) == 0)
    return (gpointer)1;
  else
    return NULL;
}

static chr *
string_matcher_subst (mptn_t *mptn, mptn_var_t *vars, gpointer data)
{
  MPTN_CHECK (mptn);
  g_assert (mptn->ops == &mptn_string_matcher_ops);

  return chra_dup ((chr*) mptn->mctl);
}

static void 
string_matcher_print (mptn_t *mptn, FILE *fp, guint offset)
{
  MPTN_CHECK (mptn);
  g_assert (mptn->ops == &mptn_string_matcher_ops);

  make_offset (fp, offset);

  fputs ("String(", fp);
  fputs (mptn->mctl, fp);	/* !!! Need to do some backslash screening -- Goga */
  fputs (")\n", fp);
}

/******************* A matcher that agrees with anything **********************/
static gpointer any_matcher_iter_start (mptn_t *mptn,
					chr *begin, chr *end,
					mptn_var_t *vars, guint stage,
					gpointer data);
static void any_matcher_print (mptn_t *mptn, FILE *fp, guint offset);


mptn_matcher_ops_t mptn_any_matcher_ops = {
  NULL,				/* No public initializer */
  NULL,				/* No public destructor */
  any_matcher_iter_start,
  NULL,				/* Since we can only have one match, 
				   no need for iter_step. */
  NULL,				/* Since we allocate no memory, 
				   no need for iter_free.*/
  NULL,				/* No stage changes here, 
				   thus no stage function. */
  NULL,				/* Impossible to recover the string,
				   no Subst function */
  any_matcher_print
};

mptn_t *
mptn_any_matcher_init (mptn_control_t *control, 
		       gboolean empty_allowed, gboolean repetition)
{
  mptn_t *res;

  control = mptn_control_normalize (control);

  res = mptn_new (control);

  res->ops = &mptn_any_matcher_ops;
  res->mctl =  NULL;
  res->min_len = empty_allowed ? 0 : 1;
  res->max_len = repetition ? G_MAXINT: 1;

  return res;
}

static gpointer 
any_matcher_iter_start (mptn_t *mptn,
			chr *begin, chr *end,
			mptn_var_t *vars, guint stage,
			gpointer data)
{
  return (gpointer)1;
}

static void 
any_matcher_print (mptn_t *mptn, FILE *fp, guint offset)
{
  MPTN_CHECK (mptn);

  make_offset (fp, offset);
  fprintf (fp, "Any(%s, %s)\n",
	   mptn->max_len == G_MAXINT ? "Rep" : "NoRep",
	   mptn->min_len == 0 ? "EmptyAllowed" : "EmptyNotAllowed");
}

/************************* An alternative *************************************/

/*
  The structure holding the alternative-specific iterator data.
*/
#define ALT_ICTL_MAGIC	0x01225408

typedef struct alt_ictl {
  guint magic;
  mptn_t **alts;	/* The Mptn for the alternative */
  chr *begin;		/* The string for matching */
  chr *end;
  mptn_var_t *vars;	/* variables to fill */
  guint stage;		/* The stage to start from */
  gpointer data;	/* Additional data */

  mptn_t **cur;		/* The current alternative */
  mptn_ectl_t *ectl;	/* Ectl for the current alternative */
} alt_ictl_t;

#define ALT_ICTL_CHECK(ic)			\
	do {					\
	  if ((ic)->magic != ALT_ICTL_MAGIC)	\
	    g_error ("Not an alt_ictl");	\
	} while(0)

static void alt_matcher_destroy (gpointer mctl);
static gpointer alt_matcher_iter_start (mptn_t *mptn,
					chr *begin, chr *end,
					mptn_var_t *vars, guint stage,
					gpointer data);
static gpointer alt_matcher_iter_step (gpointer mictl);
static void alt_matcher_ictl_free (gpointer mictl);
static guint alt_matcher_stage (gpointer mictl);
static chr *alt_matcher_subst (mptn_t *mptn, mptn_var_t *vars, gpointer data);
static void alt_matcher_print (mptn_t *mptn, FILE *fp, guint offset);

/*
  The functions table for Alternative matcher.
*/
mptn_matcher_ops_t mptn_alt_matcher_ops = {
  NULL,			/* No initialization of the standard kind */
  alt_matcher_destroy,
  alt_matcher_iter_start,
  alt_matcher_iter_step,
  alt_matcher_ictl_free,
  alt_matcher_stage,
  alt_matcher_subst,
  alt_matcher_print
};

/* 
   This matcher is generated by the compiler and does not 
   have an initialization from a string.

   The initializer it does have passes the set of alternatives.
   The whole array is copied and returned as res->mctl.
*/

mptn_t * 
mptn_alt_matcher_init (mptn_control_t *control,  mptn_t **alts)
{
  mptn_t *res;
  int nalt;	/* Number of alternatives */
  mptn_t **alts_copy;
  mptn_t **mp, **mpc;

  g_assert (alts != NULL);

  res = mptn_new (control);
  res->ops = &mptn_alt_matcher_ops;
  
  for (nalt = 0, mp = alts; *mp; nalt++, mp++)
    ;

  alts_copy = g_malloc ((nalt+1) * sizeof (mptn_t*));
  for (mp = alts, mpc = alts_copy; *mp; mp++, mpc++)
    *mpc = *mp;

  *mpc = NULL; 	/* Zero-terminate the array */

  /* Take responsibility for the mptns */
  for (mpc = alts_copy; *mpc; mpc++)
    mptn_refinc (*mpc);

  res->mctl = alts_copy;

  /* Set the length fields of the result */
  res->min_len = G_MAXINT;
  res->max_len = 0;

  for (mpc = alts_copy; *mpc; mpc++)
    {
      if ((*mpc)->min_len < res->min_len)
	res->min_len = (*mpc)->min_len;

      if ((*mpc)->max_len > res->max_len)
	res->max_len = (*mpc)->max_len;
    }

  return res;
}

static void
alt_matcher_destroy (gpointer mctl)
{
  mptn_t **alts = mctl;
  mptn_t **mp;

  g_assert (mctl != NULL);

  /* Free the alternatives */
  for (mp = alts; *mp; mp++)
    mptn_free (*mp);
  
  /* Now free the array */
  g_free (alts);
}

static gpointer
alt_matcher_iter_start (mptn_t *mptn,
			chr *begin, chr *end,
			mptn_var_t *vars, guint stage,
			gpointer data)
{
  alt_ictl_t *res;
  mptn_t **cur;
  mptn_ectl_t *ectl;

  MPTN_CHECK (mptn);
  g_assert (begin != NULL);
  g_assert (end != NULL);

  res = g_malloc (sizeof (alt_ictl_t));
  res->magic = ALT_ICTL_MAGIC;
  res->alts = mptn->mctl;
  res->begin = begin;
  res->end = end;
  res->vars = vars;
  res->stage = stage;		/* Subpatterns will be applied at the next stage */
  res->data = data;

  for (cur = mptn->mctl; *cur; cur++)
    {
      ectl = mptn_exec_start_with_vars (*cur, begin, end, 
					vars, stage+1, 
					data);
      if (ectl)
	break;
    }

  if (ectl) 	/* Some match found */
    {
      res->cur = cur;
      res->ectl = ectl;
      return res;
    }
  else		/* No match worked */
    {	/* Need to destroy Res */
      g_free (res);
      return NULL;
    }
}

static gpointer
alt_matcher_iter_step (gpointer mictl)
{
  alt_ictl_t *aictl = mictl;
  
  ALT_ICTL_CHECK (aictl);

  aictl->ectl = mptn_exec_step (aictl->ectl);
  
  if (aictl->ectl)	/* The current alternative still works */
    return aictl;
  else			/* The current alternative is exhausted */
    {
      mptn_t **cur;
      mptn_ectl_t *ectl;

      ectl = NULL;
      for (cur = aictl->cur + 1; *cur; cur++)
	{
	  ectl = mptn_exec_start_with_vars (*cur, 
					    aictl->begin, aictl->end, 
					    aictl->vars, 
					    aictl->stage+1, 
					    aictl->data);
	  if (ectl)
	    break;
	}

      if (ectl) 	/* Some other alternative is found */
	{
	  aictl->cur = cur;
	  aictl->ectl = ectl;
	  return aictl;
	}
      else		/* No more alternatives */
	{
	  g_free (aictl);
	  return NULL;
	}
	 
    }
}

static void
alt_matcher_ictl_free (gpointer mictl)
{
  alt_ictl_t *aictl = mictl;

  ALT_ICTL_CHECK (aictl);
  if (aictl->ectl)
    mptn_ectl_free (aictl->ectl);
  g_free (aictl);
}

static guint
alt_matcher_stage (gpointer mictl)
{
  alt_ictl_t *aictl = mictl;

  ALT_ICTL_CHECK (aictl);
  return (mptn_ectl_stage (aictl->ectl));
}

static chr *
alt_matcher_subst (mptn_t *mptn, mptn_var_t *vars, gpointer data)
{
  mptn_t **cur;
  chr * res;

  MPTN_CHECK (mptn);

  /* This function works for Alt and Conj */
  g_assert (mptn->ops == &mptn_alt_matcher_ops || mptn->ops == &mptn_conj_matcher_ops);

  /* Return the first variant that goes */
  for (cur = (mptn_t**) mptn->mctl; *cur; cur++)
    if ((res = mptn_subst (*cur, vars, data)))
      return res;

  return NULL;
}

static void 
alt_matcher_print (mptn_t *mptn, FILE *fp, guint offset)
{
  mptn_t **alts;

  MPTN_CHECK (mptn);
  g_assert (fp != NULL);

  make_offset (fp, offset);
  fputs ("Alt\n", fp);

  for (alts = mptn->mctl; *alts; alts++)
    mptn_print (*alts, fp, offset + 2);

}

/************************** A conjunction *************************************/

/*
  A structure holding the conjunction specific iterator data.
*/
#define CONJ_ICTL_MAGIC 0x03321148

typedef struct conj_ictl_t {
  guint magic;
  guint n_sub;		/* Number of conjuncts */
  mptn_t **subs;	/* The Mptns for the conjuncted expressions */
  mptn_ectl_t **ectls;	/* Ectls for them */
  guint	*stages;	/* Stages the various expressions are at .
			   Stages[0] is the initial stage passed to the iterator.
			   Stages[i+1] is the stage returned by i-th subiterator.
			 */
  chr *begin;		/* The string for matching */
  chr *end;
  mptn_var_t *vars;	/* variables to fill */
  gpointer data;	/* Additional data */
} conj_ictl_t;

#define CONJ_ICTL_CHECK(ic)			\
	do {					\
	  if ((ic)->magic != CONJ_ICTL_MAGIC)	\
	    g_error ("Not an conj_ictl");	\
	} while(0)

static void conj_matcher_destroy (gpointer mctl);
static gboolean conj_iter_start_at_level (conj_ictl_t *cictl, guint level);
static gboolean conj_iter_step_at_level (conj_ictl_t *cictl, guint level);
static gpointer conj_matcher_iter_start (mptn_t *mptn,
					 chr *begin, chr *end,
					 mptn_var_t *vars, guint stage,
					 gpointer data);
static gpointer conj_matcher_iter_step (gpointer mictl);
static void conj_matcher_ictl_free (gpointer ictl);
static guint conj_matcher_stage (gpointer mictl);
static void conj_matcher_print (mptn_t *mptn, FILE *fp, guint offset);

/*
  The functions table for Conjunction matcher.
*/
mptn_matcher_ops_t mptn_conj_matcher_ops = {
  NULL,			/* No initialization of the standard kind */
  conj_matcher_destroy,
  conj_matcher_iter_start,
  conj_matcher_iter_step,
  conj_matcher_ictl_free,
  conj_matcher_stage,
  alt_matcher_subst, 	/* Subst is the same for Alt and Conj */
  conj_matcher_print
};

/* 
   This matcher is generated by the compiler and does not 
   have an initialization from a string.

   The initializer it does have passes the set of subexpressionl.
   The whole array is copied and returned as res->mctl.
   (It would make sense to also keep the number of expressions, but I decided
   it's easier to compute them every time a matching starts -- not such a big work 
   -- Goga)
*/

mptn_t * 
mptn_conj_matcher_init (mptn_control_t *control,  mptn_t **subs)
{
  mptn_t *res;
  int nsub;	/* Number of subexpressions */
  mptn_t **subs_copy;
  mptn_t **mp, **mpc;

  g_assert (subs != NULL);

  res = mptn_new (control);
  res->ops = &mptn_conj_matcher_ops;
  
  for (nsub = 0, mp = subs; *mp; nsub++, mp++)
    ;

  subs_copy = g_malloc ((nsub+1) * sizeof (mptn_t*));
  for (mp = subs, mpc = subs_copy; *mp; mp++, mpc++)
    *mpc = *mp;

  *mpc = NULL; 	/* Zero-terminate the array */

  /* Take responsibility for the mptns */
  for (mpc = subs_copy; *mpc; mpc++)
    mptn_refinc (*mpc);

  res->mctl = subs_copy;

  /* Set the length fields of the result */
  /* 
     This computation (and the ops field assignment) are the only things
     that keep this procedure distinct from Mptn_alt_matcher_init -- Goga
  */
  res->min_len = 0;		
  res->max_len = G_MAXINT;

  for (mpc = subs_copy; *mpc; mpc++)
    {
      if ((*mpc)->min_len > res->min_len)
	res->min_len = (*mpc)->min_len;

      if ((*mpc)->max_len < res->max_len)
	res->max_len = (*mpc)->max_len;
    }

  return res;
}

static void
conj_matcher_destroy (gpointer mctl)
{
  mptn_t **subs = mctl;
  mptn_t **mp;

  g_assert (subs != NULL);

  /* Free the alternatives */
  for (mp = subs; *mp; mp++)
    mptn_free (*mp);
  
  /* Now free the array */
  g_free (subs);
}

/*
  Start an iterator for Conj at Level.
*/
static gboolean
conj_iter_start_at_level (conj_ictl_t *cictl, guint level)
{
  CONJ_ICTL_CHECK (cictl);

  /* 
     Try to find a state of the iterator where all the higher levels
     get satisfied.
  */
  for (cictl->ectls[level] = mptn_exec_start_with_vars (cictl->subs[level],
							cictl->begin, cictl->end,
							cictl->vars,
							cictl->stages[level] + 1,
							cictl->data);
       cictl->ectls[level];
       cictl->ectls[level] = mptn_exec_step (cictl->ectls[level]))
    {
      cictl->stages[level+1] = mptn_ectl_stage (cictl->ectls[level]);
      if (level == cictl->n_sub - 1 	/* Reached the last one */
	  || conj_iter_start_at_level (cictl, level + 1))
	return TRUE;
    }
  /* 
     Dropped out of the loop,
     could not start the iterator at this level.
  */
  return FALSE;
}

/*
  Make an iterator step at Level.
*/
static gboolean
conj_iter_step_at_level (conj_ictl_t *cictl, guint level)
{
  CONJ_ICTL_CHECK (cictl);

  /*
    Step the iterator until we are able to start the higher levels.
  */
  while ((cictl->ectls[level] = mptn_exec_step (cictl->ectls[level])))
    {
      cictl->stages[level+1] = mptn_ectl_stage (cictl->ectls[level]);
      if (level == cictl->n_sub - 1    				/* The last level */
	  || conj_iter_start_at_level (cictl, level+1) )	/* Higher levels start successfully */
	return TRUE;
    }

  /* Could not advance this iterator and satisfy the higher levels */
  return FALSE;
}

static gpointer
conj_matcher_iter_start (mptn_t *mptn,
			 chr *begin, chr *end,
			 mptn_var_t *vars, guint stage,
			 gpointer data)
{
  conj_ictl_t *res;
  int n_sub;
  mptn_t **ms;

  MPTN_CHECK (mptn);
  g_assert (begin != NULL);
  g_assert (end != NULL);

  /* First count the number of subexpressions */
  for (n_sub = 0, ms = mptn->mctl; *ms; n_sub++, ms++)
    ;

  res = g_malloc (sizeof (conj_ictl_t));
  res->magic = CONJ_ICTL_MAGIC;
  res->n_sub = n_sub;
  res->subs = mptn->mctl;
  res->ectls = g_new0 (mptn_ectl_t*, n_sub);
  res->stages = g_new0 (guint, n_sub + 1);
  res->stages[0] = stage;
  res->begin = begin;
  res->end = end;
  res->vars = vars;
  res->data = data;

  if (conj_iter_start_at_level (res, 0))
    return res;
  else
    {
      conj_matcher_ictl_free (res);
      return NULL;
    }
}

static gpointer
conj_matcher_iter_step (gpointer mictl)
{
  conj_ictl_t *cictl = mictl;
  int i;
  
  CONJ_ICTL_CHECK (cictl);

  for (i = cictl->n_sub - 1; i >= 0; i--)
    if (conj_iter_step_at_level (cictl, i)) 	
      /* Backtracked to level i, 
	 alright there, returning */
      return cictl;
  
  /* Down to level 0 could not find an iterator to advance */
  conj_matcher_ictl_free (cictl);
  return NULL;
}

static void
conj_matcher_ictl_free (gpointer mictl)
{
  conj_ictl_t *cictl = mictl;
  int i;

  CONJ_ICTL_CHECK (cictl);

  for (i = 0; i < cictl->n_sub; i++)
    if (cictl->ectls[i])
      mptn_ectl_free (cictl->ectls[i]);

  g_free (cictl->ectls);
  g_free (cictl->stages);

  g_free (cictl);
}

static guint
conj_matcher_stage (gpointer mictl)
{
  conj_ictl_t *cictl = mictl;

  CONJ_ICTL_CHECK (cictl);
  return cictl->stages [cictl->n_sub];
}

static void 
conj_matcher_print (mptn_t *mptn, FILE *fp, guint offset)
{
  mptn_t **subs;

  MPTN_CHECK (mptn);
  g_assert (fp != NULL);

  make_offset (fp, offset);
  fputs ("Conj\n", fp);

  for (subs = mptn->mctl; *subs; subs++)
    mptn_print (*subs, fp, offset + 2);
}

/************************** Concatenation *************************************/

/*
  A structure holding the concatenation expression data.
*/
#define CAT_MCTL_MAGIC 0x04333806

typedef struct cat_mctl {
  guint magic;
  guint n_sub;		/* Number of subexpressions */
  mptn_t **subs;	/* The Mptns for the concatenated subexpressions */
  guint nf_beg;		/* Number of fixed length subexpresions at the beginning */
  guint nf_end;		/* Number of fixed length subexpressions at the end */
  guint *flags;		/* Direction of trial for borders for subexpressions.
			   (maybe later there will be more) */
  guint *slack;		/* How many more characters than minimum can the rest of
			   the expression accept */
  guint fixed_at_left;	/* Total fixed length part at the left */
  gint frozen_point;	/* Below this level everything is matched only once */
} cat_mctl_t;

#define CAT_MCTL_CHECK(ic)			\
	do {					\
	  if ((ic)->magic != CAT_MCTL_MAGIC)	\
	    g_error ("Not a cat_mctl");	\
	} while(0)

/*
  A structure holding the concatenation specific iterator data.
*/
#define CAT_ICTL_MAGIC 0x04790424

typedef struct cat_ictl {
  guint magic;
  cat_mctl_t *mctl;	/* Expression information */
  mptn_ectl_t **ectls;	/* Ectls for subexpressions */
  guint	*stages;	/* Stages the various expressions are at .
			   Stages[0] is the initial stage passed to the iterator.
			   Stages[i+1] is the stage returned by i-th subiterator.
			 */
  guint *off_left;	/* Offset for the left border of subexpressions
			  -- array by levels
			  (some tricky computations required when jumping from 
			  beginning to end and the to varlen).
			*/
  guint *slack_rem;	/* How much characters should we be able to cover above 
			   the minimal possible length
			   -- array by levels.
			 */
  chr *begin;		/* The string for matching */
  chr *end;
  mptn_var_t *vars;	/* variables to fill */
  gpointer data;	/* Additional data */
} cat_ictl_t;

#define CAT_ICTL_CHECK(ic)			\
	do {					\
	  if ((ic)->magic != CAT_ICTL_MAGIC)	\
	    g_error ("Not a cat_ictl");	\
	} while(0)

static void cat_matcher_destroy (gpointer mctl);
static gboolean cat_iter_start_at_level (cat_ictl_t *cictl, guint level);
static gboolean cat_iter_step_at_level (cat_ictl_t *cictl, guint level);
static gboolean cat_iter_start_with_level_and_slack (cat_ictl_t *cictl,
						     guint level,
						     guint slack);
static gpointer cat_matcher_iter_start (mptn_t *mptn,
					 chr *begin, chr *end,
					 mptn_var_t *vars, guint stage,
					 gpointer data);
static gpointer cat_matcher_iter_step (gpointer mictl);
static void cat_matcher_ictl_free (gpointer ictl);
static guint cat_matcher_stage (gpointer mictl);
static chr *cat_matcher_subst (mptn_t *mptn, mptn_var_t *vars, gpointer data);
static void cat_matcher_print (mptn_t *mptn, FILE *fp, guint offset);

/*
  The functions table for Concatenation matcher.
*/
mptn_matcher_ops_t mptn_cat_matcher_ops = {
  NULL,			/* No initialization of the standard kind */
  cat_matcher_destroy,
  cat_matcher_iter_start,
  cat_matcher_iter_step,
  cat_matcher_ictl_free,
  cat_matcher_stage,
  cat_matcher_subst,
  cat_matcher_print
};

/* 
   This matcher is generated by the compiler and does not 
   have an initialization from a string.

   The initializer it does have passes the set of subexpressions.
   The whole array is copied and returned as a part of res->mctl.
*/
mptn_t * 
mptn_cat_matcher_init (mptn_control_t *control,  mptn_t **subs)
{
  mptn_t *res;
  mptn_t **mp, **mpc;
  cat_mctl_t *cmctl;
  mptn_t **subs_copy;		/* Values for fields in Cmctl */
  int n_sub;	/* Number of subexpressions */
  guint nf_beg, nf_end;
  guint *flags;
  guint *slack;
  guint fixed_at_left;
  gint frozen_point;

  g_assert (subs != NULL);

  res = mptn_new (control);
  res->ops = &mptn_cat_matcher_ops;
  
  for (n_sub = 0, mp = subs; *mp; n_sub++, mp++)
    ;

  subs_copy = g_new0 (mptn_t*, n_sub + 1);

  /* 
     Copy the fixed length subexpressions 
     at the beginning of concatenation
     (these will be matched first).
  */
  for (mp = subs, mpc = subs_copy, nf_beg = 0; 
       *mp && ((*mp)->min_len == (*mp)->max_len); 
       mp++, mpc++, nf_beg++)
    *mpc = *mp;

  if (nf_beg == n_sub) /* there are no variable length subexpressions */
    nf_end = 0;
  else			/* there are some */
    {
      int left_to_copy;

      /* mpc is already pointing at the end of the filled part of subs_copy. */
      /* 
	 Now copy the fixed length subexpressions at the end of concatenation 
	 (these will be matched after the ones at the beginning)
      */
      for (mp = subs + n_sub - 1, nf_end = 0;
	   (*mp)->min_len == (*mp)->max_len;
	   mp--, mpc++, nf_end++)
	*mpc = *mp;
      
      /* There should be something in the middle */
      g_assert (mp >= subs);
      g_assert (nf_beg + nf_end < n_sub);

      /* mpc still points at the end of the copied part */
      /* 
	 Finally, copy the middle part containing 
	 the variable length subexpressions 
	 (these will be the last to match)
      */

      for (mp = subs + nf_beg,
	     left_to_copy = n_sub - nf_beg - nf_end;
	   left_to_copy;
	   mp++, mpc++, left_to_copy--)
	*mpc = *mp;
    }

  *mpc = NULL; 	/* Zero-terminate the array */

  /* Take responsibility for the mptns */
  for (mpc = subs_copy; *mpc; mpc++)
    mptn_refinc (*mpc);

  /* Compute the flags fields in the Mctl */
  flags = g_new0 (guint, n_sub);
  
  /* The first (mctl->nf_beg + mtcl->nf_end) elements have zero flags */

  /* Compute flags and frozen_point */
  /* Frozen point is the last level with nonzero flags +1
     or 0 if no flags present */
  {
    guint flpoint; 	/* Starting from where do we have similar flags? */
    guint i, j;

    frozen_point = 0;
    flpoint = nf_beg + nf_end;
    for (i = nf_beg + nf_end;
	 i < n_sub;
	 i++)
      if (subs_copy[i]->flags)
	{
	  /* Found a subexpression with nonzero flags */
	  /* 
	     Copy these flags for all the past subexpressions starting
	     from flpoint.
	  */
	  for (j = flpoint; j <= i; j++)
	    flags[j] = subs_copy[i]->flags;
	  frozen_point = i+1;
	  flpoint = i+1;	/* The next group starts */
	}
	
  }

  /* Set the length fields of the result */

  /* Minimal length is just the sum of subexpressions' minimal lengths */
  res->min_len = 0;		
  for (mpc = subs_copy; *mpc; mpc++)
      res->min_len += (*mpc)->min_len;

  /* 
     When computing the maximal length, we need to take into account
     that some subexpression may have arbitrary big length.
  */
  res->max_len = 0;
  for (mpc = subs_copy; *mpc; mpc++)
    if ((*mpc)->max_len == G_MAXINT)
      {
	res->max_len = G_MAXINT;
	break;
      }
    else
      res->max_len += (*mpc)->max_len;

  /* Compute the slack */
  {
    gint i;
    
    slack = g_new0 (guint, n_sub+1);
    slack[n_sub] = 0;
    for (i = n_sub-1; i >= 0; i--)
      {
	if (subs_copy[i]->max_len == G_MAXINT
	    || slack[i+1] == G_MAXINT)
	  slack[i] = G_MAXINT;
	else
	  slack[i] = slack[i+1] 
	    + subs_copy[i]->max_len - subs_copy[i]->min_len;
      }
  }

  /* Compute fixed_at_left */
  {
    int i;
    fixed_at_left = 0;
    for (i = 0, mp = subs_copy; i < nf_beg; i++, mp++)
      fixed_at_left += (*mp)->min_len;
  }

  cmctl = g_new0 (cat_mctl_t, 1);
  cmctl->magic = CAT_MCTL_MAGIC;
  cmctl->n_sub = n_sub;
  cmctl->subs = subs_copy;
  cmctl->nf_beg = nf_beg;
  cmctl->nf_end = nf_end;
  cmctl->flags = flags;
  cmctl->slack = slack;
  cmctl->fixed_at_left = fixed_at_left;
  cmctl->frozen_point = frozen_point;

  res->mctl = cmctl;

  return res;
}

/*
  Free the memory occupied by Cmctl and referred by it.
*/
static void
cat_matcher_destroy (gpointer mctl)
{
  cat_mctl_t *cmctl = mctl;
  int i;

  CAT_MCTL_CHECK (cmctl);

  for (i = 0; i < cmctl->n_sub; i++)
    mptn_free (cmctl->subs[i]);
  g_free (cmctl->subs);

  g_free (cmctl->flags);
  g_free (cmctl->slack);

  g_free (cmctl);
}

static gboolean
cat_iter_start_at_level (cat_ictl_t *cictl, guint level)
{
  cat_mctl_t *cmctl;
  mptn_t *subex;	/* Subexpression we are going to work on */
  gint max_slack, min_slack;

  CAT_ICTL_CHECK (cictl);

  cmctl = cictl->mctl;
  CAT_MCTL_CHECK (cmctl);
  subex = cmctl->subs[level];

  /* Find which part of the string we are operationg on */

  if (level < cmctl->nf_beg + cmctl->nf_end)
    /* Fixed place*/
    {
      g_assert (subex->min_len == subex->max_len);
    }
  else if (level == cmctl->nf_beg + cmctl->nf_end)
    /* Variable length part of the expression starts */
    {
      g_assert (subex->min_len < subex->max_len); 	/* Should be var length */
    }

  /* Maximum extra chars we can eat */
  max_slack = MIN (cictl->slack_rem[level], 
		   subex->max_len - subex->min_len);
  /* Minimum extra chars we are allowed to eat */
  if (cictl->slack_rem[level] < cmctl->slack[level+1])
    min_slack = 0;
  else
    min_slack = cictl->slack_rem[level] - cmctl->slack[level+1];

  /* For fixed length expressions, both should be zero */
  g_assert (subex->max_len == subex->min_len
	    ? max_slack == 0 && min_slack == 0
	    : min_slack <= max_slack);

  /* Now start the loop itself; the special case is MPTN_LONGER */
  if (cmctl->flags[level] & MPTN_LONGER)
    /* Trying starts from the longest variant */
    {
      int slack;

      for (slack = max_slack; slack >= min_slack; slack--)
	if (cat_iter_start_with_level_and_slack (cictl, level, slack))
	  return TRUE;
      return FALSE; 	/* Tried all alternatives, failed */
    }
  else /* not MPTN_LONGER -- trying starts from the shortest variant */
    {
      int slack;

      for (slack = min_slack; slack <= max_slack; slack++)
	if (cat_iter_start_with_level_and_slack (cictl, level, slack))
	  return TRUE;
      return FALSE; 	/* Tried all alternatives, failed */
    }
}

static gboolean 
cat_iter_step_at_level (cat_ictl_t *cictl, guint level)
{
  cat_mctl_t *cmctl;
  mptn_t *subex;

  CAT_ICTL_CHECK (cictl);

  cmctl = cictl->mctl;
  CAT_MCTL_CHECK (cmctl);

  subex = cmctl->subs[level];
  MPTN_CHECK (subex);

  if (level < cmctl->frozen_point) /* No other variants to be looked at here */
    {
      mptn_ectl_free (cictl->ectls[level]);	/* No longer needed */
      cictl->ectls[level] = NULL;
      return FALSE;
    }

  /* 
     Step the iterator until we are able to start the higher levels, 
     without rearranging here.
  */
  while ((cictl->ectls[level] = mptn_exec_step (cictl->ectls[level])))
    {
      cictl->stages[level+1] = mptn_ectl_stage (cictl->ectls[level]);
      if (level == cmctl->n_sub - 1    				/* The last level */
	  || cat_iter_start_at_level (cictl, level+1) )	/* Higher levels start successfully */
	return TRUE;
    }

  /* Try to advance the border here and restart higher levels */
  /* This can only happen when no flags are set; otherwise we would be frozen;
     therefore, we may only increase the slack */
  { 
    guint max_slack;
    gint cur_slack;
    guint slack;

    /* Maximum extra chars we can eat */
    max_slack = MIN (cictl->slack_rem[level], 
		     subex->max_len - subex->min_len);

    /* The slack we had previously */
    cur_slack = cictl->slack_rem[level] - cictl->slack_rem[level+1];
  
    for (slack = cur_slack+1; slack <= max_slack; slack++)
      if (cat_iter_start_with_level_and_slack (cictl, level, slack))
	return TRUE;
  }
  return FALSE; 	/* Tried all alternatives, failed */
}

static gboolean
cat_iter_start_with_level_and_slack (cat_ictl_t *cictl,
				     guint level,
				     guint slack)
{
  cat_mctl_t *cmctl;
  mptn_t *subex;

  CAT_ICTL_CHECK (cictl);
  cmctl = cictl->mctl;
  CAT_MCTL_CHECK (cmctl);
  g_assert (level < cmctl->n_sub);

  subex = cmctl->subs[level];

  if (level >= cmctl->nf_beg + cmctl->nf_end)	/* Dynamic computation of off_left */
    cictl->off_left[level+1] = cictl->off_left[level] + subex->min_len + slack;
  cictl->slack_rem[level+1] = cictl->slack_rem[level] - slack;
  
  for (cictl->ectls[level] = mptn_exec_start_with_vars (subex,
							cictl->begin + cictl->off_left[level],
							cictl->begin + cictl->off_left[level]
							+ subex->min_len + slack,
							cictl->vars,
							cictl->stages[level] + 1,
							cictl->data);
       cictl->ectls[level];
       cictl->ectls[level] = mptn_exec_step (cictl->ectls[level]))
    {
      cictl->stages[level+1] = mptn_ectl_stage (cictl->ectls[level]);
      if (level == cmctl->n_sub - 1 	/* Reached the last one */
	  || cat_iter_start_at_level (cictl, level + 1))
	return TRUE;	/* All right, got to the end of matching */
    }
  return FALSE;
}

static gpointer 
cat_matcher_iter_start (mptn_t *mptn,
			chr *begin, chr *end,
			mptn_var_t *vars, guint stage,
			gpointer data)
{
  cat_ictl_t *res;
  cat_mctl_t *cmctl;

  MPTN_CHECK (mptn);
  g_assert (begin != NULL);
  g_assert (end != NULL);

  cmctl = mptn->mctl;
  CAT_MCTL_CHECK (cmctl);

  res = g_new0 (cat_ictl_t, 1);
  res->magic = CAT_ICTL_MAGIC;
  res->mctl = cmctl;
  res->ectls = g_new0 (mptn_ectl_t*, cmctl->n_sub);
  
  res->stages = g_new0 (guint, cmctl->n_sub + 1);
  res->stages[0] = stage;

  res->off_left = g_new0 (guint, cmctl->n_sub + 1);

  /* Static computation of off_left */
  {
    int i;

    res->off_left[0] = 0;
    for (i = 1; i < cmctl->nf_beg; i++)
      res->off_left[i] = res->off_left[i-1] + cmctl->subs[i-1]->min_len;

    if (cmctl->nf_end > 0)
      {
	res->off_left [cmctl->nf_beg] = end - begin - cmctl->subs[cmctl->nf_beg]->min_len;
	for (i = cmctl->nf_beg + 1; i < cmctl->nf_beg + cmctl->nf_end; i++)
	  res->off_left[i] = res->off_left[i-1] - cmctl->subs[i]->min_len;
      }

    res->off_left[cmctl->nf_beg + cmctl->nf_end] = cmctl->fixed_at_left;
  }

  res->slack_rem = g_new0 (guint, cmctl->n_sub + 1);
  res->slack_rem[0] = end - begin - mptn->min_len;
  
  res->begin = begin;
  res->end = end;
  res->vars = vars;
  res->data = data;

  if (cat_iter_start_at_level (res, 0))
    return res;
  else
    {
      cat_matcher_ictl_free (res);
      return NULL;
    }
}

static gpointer
cat_matcher_iter_step (gpointer mictl)
{
  cat_ictl_t *cictl = mictl;
  int i;
  cat_mctl_t *cmctl;
  
  CAT_ICTL_CHECK (cictl);
  cmctl = cictl->mctl;
  CAT_MCTL_CHECK (cmctl);

  for (i = cmctl->n_sub - 1; i >= 0; i--)
    if (cat_iter_step_at_level (cictl, i)) 	
      /* Backtracked to level i, 
	 alright there, returning */
      return cictl;
  
  /* Down to level 0 could not find an iterator to advance */
  cat_matcher_ictl_free (cictl);
  return NULL;
}

static void
cat_matcher_ictl_free (gpointer mictl)
{
  cat_ictl_t *cictl = mictl;
  int i;
  cat_mctl_t *cmctl;

  CAT_ICTL_CHECK (cictl);
  cmctl= cictl->mctl;
  CAT_MCTL_CHECK (cmctl);

  for (i = 0; i < cmctl->n_sub; i++)
    if (cictl->ectls[i])
      mptn_ectl_free (cictl->ectls[i]);

  g_free (cictl->ectls);
  g_free (cictl->stages);
  g_free (cictl->off_left);
  g_free (cictl->slack_rem);

  g_free (cictl);
}

static guint
cat_matcher_stage (gpointer mictl)
{
  cat_ictl_t *cictl = mictl;
  cat_mctl_t *cmctl;

  CAT_ICTL_CHECK (cictl);
  cmctl = cictl->mctl;
  CAT_MCTL_CHECK (cmctl);

  return cictl->stages [cmctl->n_sub];
}

/*
  !!!!! This function only works for chr == char -- Goga
*/
static chr *
cat_matcher_subst (mptn_t *mptn, mptn_var_t *vars, gpointer data)
{
  cat_mctl_t *cmctl;
  GString *gstr;
  chr *sres;
  chr *res;
  guint i;

  MPTN_CHECK (mptn);
  
  cmctl = (cat_mctl_t*) mptn->mctl;
  CAT_MCTL_CHECK (cmctl);

  gstr = g_string_new ("");

  /* !!!!! Is there a way to express the loop 
     without repeating the same chunk of code three times? -- Goga */

  /* First, (try to) append the fixed length subexpressions at
     the beginning */
  for (i = 0; i < cmctl->nf_beg; i++)
    {
      sres = mptn_subst (cmctl->subs[i], vars, data);
      if (!sres)
	{
	  g_string_free (gstr, TRUE);
	  return NULL;
	}
      else
	{
	  g_string_append (gstr, sres);
	  g_free (sres);
	}
    }

  /* Then, variable length */
  for (i = cmctl->nf_beg + cmctl->nf_end; i < cmctl->n_sub; i++)
    {
      sres = mptn_subst (cmctl->subs[i], vars, data);
      if (!sres)
	{
	  g_string_free (gstr, TRUE);
	  return NULL;
	}
      else
	{
	  g_string_append (gstr, sres);
	  g_free (sres);
	}
    }

  /* Finally, fixed length subexpressions at the end */
  for (i = cmctl->nf_beg; i < cmctl->nf_beg + cmctl->nf_end; i++)
    {
      sres = mptn_subst (cmctl->subs[i], vars, data);
      if (!sres)
	{
	  g_string_free (gstr, TRUE);
	  return NULL;
	}
      else
	{
	  g_string_append (gstr, sres);
	  g_free (sres);
	}
    }

  res = gstr->str;
  g_string_free (gstr, FALSE);

  return res;
}

static void 
cat_matcher_print (mptn_t *mptn, FILE *fp, guint offset)
{
  int i;
  cat_mctl_t *cmctl;

  MPTN_CHECK (mptn);
  g_assert (fp != NULL);

  cmctl= mptn->mctl;
  CAT_MCTL_CHECK (cmctl);

  make_offset (fp, offset);
  fputs ("Cat\n", fp);

  if (cmctl->nf_beg != 0)
    {
      make_offset (fp, offset+1);
      fputs ("Begin:\n", fp);
      for (i = 0; i < cmctl->nf_beg; i++)
	mptn_print (cmctl->subs[i], fp, offset+2);
    }
  if (cmctl->nf_end != 0)
    {
      make_offset (fp, offset+1);
      fputs ("End:\n", fp);
      for (i = cmctl->nf_beg; i < cmctl->nf_beg + cmctl->nf_end; i++)
	mptn_print (cmctl->subs[i], fp, offset+2);
    }
  if (cmctl->nf_beg + cmctl->nf_end < cmctl->n_sub)
    {
      make_offset (fp, offset+1);
      fputs ("Variable length:\n", fp);
      for (i = cmctl->nf_beg + cmctl->nf_end; i < cmctl->n_sub; i++)
	mptn_print (cmctl->subs[i], fp, offset+2);
    }
}

/*************************** A chrset *****************************************/

static void mptn_chrset_matcher_destroy (gpointer mctl);
static gpointer mptn_chrset_matcher_iter_start (mptn_t *mptn,
					   chr *begin, chr *end,
					   mptn_var_t *vars, guint stage,
					   gpointer data);
static void mptn_chrset_matcher_print (mptn_t *mptn, FILE *fp, guint offset);

mptn_matcher_ops_t mptn_mptn_chrset_matcher_ops = {
  NULL,
  mptn_chrset_matcher_destroy,
  mptn_chrset_matcher_iter_start,
  NULL,				/* Since we can only have one match, 
				   no need for iter_step. */
  NULL,				/* Since we allocate no memory, 
				   no need for iter_free.*/
  NULL,				/* No stage changes here, 
				   thus no stage function. */
  NULL,				/* We are not able to recover
				   the string -- no Subst function */
  mptn_chrset_matcher_print
};

/*
  Initialization
*/
mptn_t *
mptn_mptn_chrset_init (mptn_control_t *control, mptn_chrset_t *chrset)
{
  mptn_t *res;

  g_assert (chrset != NULL);

  res = mptn_new (control);
  res->ops = &mptn_mptn_chrset_matcher_ops;
  res->mctl = mptn_chrset_copy (chrset);
  res->min_len = res->max_len = 1;

  return res;
}

static void 
mptn_chrset_matcher_destroy (gpointer mctl)
{
  mptn_chrset_destroy ((mptn_chrset_t*) mctl);
}

static gpointer 
mptn_chrset_matcher_iter_start (mptn_t *mptn,
			   chr *begin, chr *end,
			   mptn_var_t *vars, guint stage,
			   gpointer data)
{
  MPTN_CHECK (mptn);
  g_assert (end - begin == 1);
  
  if (mptn_chrset_in ((mptn_chrset_t*) (mptn->mctl), *begin))
    return (gpointer) 1;
  else
    return NULL;
}

static void 
mptn_chrset_matcher_print (mptn_t *mptn, FILE *fp, guint offset)
{
  guint i;
  mptn_chrset_t *cs;

  MPTN_CHECK (mptn);
  g_assert (fp != NULL);
  
  cs = mptn->mctl;

  make_offset (fp, offset);
  fputs ("Chrset[", fp);

  for (i = 0; i < CHR_NVAL; i++)
    {
      if (mptn_chrset_in (cs, i))
	{
	  if (isprint (i))
	    fputc(i, fp);	/* !!! Should do \ screening -- Goga */
	  else
	    fprintf (fp, "<0x%x>", i);
	}
    }

  fputs ("]\n", fp);
}

/*********** Repetition (also handles zero or one) *****************************/


static void rep_matcher_destroy (gpointer mctl);
static gpointer rep_matcher_iter_start (mptn_t *mptn,
					chr *begin, chr *end,
					mptn_var_t *vars, guint stage,
					gpointer data);
static gpointer rep_matcher_iter_step (gpointer mictl);
static void rep_matcher_ictl_free (gpointer mictl);
static guint rep_matcher_stage (gpointer mictl);
static chr *rep_matcher_subst (mptn_t *mptn, mptn_var_t *vars, gpointer data);
static void rep_matcher_print (mptn_t *mptn, FILE *fp, guint offset);
mptn_matcher_ops_t mptn_rep_matcher_ops = {
  NULL,
  rep_matcher_destroy,
  rep_matcher_iter_start,
  rep_matcher_iter_step,
  rep_matcher_ictl_free,
  rep_matcher_stage,
  rep_matcher_subst, 	/* only attempts to work for less-than-one rep case */
  rep_matcher_print
};

/* Structure corresponding to Mctl for Rep_matcher */
#define REP_MCTL_MAGIC 0x01174753

typedef struct rep_mctl {
  guint magic;
  mptn_t *sub;		  /* Subexpression to repeat */
  gboolean zero_allowed;  /* Are zero occurrences of subexpression allowed? */
  gboolean rep_allowed;	  /* Are repetitions allowed? */
} rep_mctl_t;

#define REP_MCTL_CHECK(ic)			\
	do {					\
	  if ((ic)->magic != REP_MCTL_MAGIC)	\
	    g_error ("Not a rep_mctl");	\
	} while(0)

/* Structure corresponding to Ictl */
typedef struct rep_ictl {
  guint magic;
  rep_mctl_t *mctl;
  chr *begin;
  chr *end;
  mptn_var_t *vars;
  guint stage;
  gpointer data;
  chr *bb;		/* The place where the second occurrence begins */
  mptn_ectl_t *ectl;	/* Ectl for the first occurrence */
  struct rep_ictl *sub_ictl; 	/* Ictl for the rest */
} rep_ictl_t;

#define REP_ICTL_MAGIC 0x01159344

#define REP_ICTL_CHECK(ic)			\
	do {					\
	  if ((ic)->magic != REP_ICTL_MAGIC)	\
	    g_error ("Not a rep_ictl");	\
	} while(0)


static rep_ictl_t *rep_sub_ictl_start (rep_ictl_t *rictl);

/*
  Initialization
*/
mptn_t *
mptn_rep_matcher_init (mptn_control_t *control, mptn_t *sub, 
		       gboolean zero_allowed, gboolean rep_allowed)
{
  mptn_t *res;
  rep_mctl_t *mctl;

  MPTN_CHECK (sub);

  res = mptn_new (control);
  res->ops = &mptn_rep_matcher_ops;

  mctl = g_new0 (rep_mctl_t, 1);
  mctl->magic = REP_MCTL_MAGIC;
  mctl->sub = sub;
  mctl->zero_allowed = zero_allowed;
  mctl->rep_allowed = rep_allowed;

  mptn_refinc (sub);	/* Take care of this */

  res->mctl = mctl;
  res->min_len = zero_allowed ? 0 : sub->min_len;
  res->max_len = (sub->max_len == 0) 
    ? 0 /* An exotic case */
    : G_MAXINT;

  return res;
}

static void 
rep_matcher_destroy (gpointer mctl)
{
  rep_mctl_t *rmctl = mctl;

  REP_MCTL_CHECK (rmctl);
  mptn_free (rmctl->sub);
  g_free (rmctl);
}

static gpointer 
rep_matcher_iter_start (mptn_t *mptn,
			chr *begin, chr *end,
			mptn_var_t *vars, guint stage,
			gpointer data)
{
  rep_mctl_t *rmctl;
  rep_ictl_t *rictl;

  MPTN_CHECK (mptn);
  g_assert (begin != NULL);
  g_assert (end != NULL);

  rmctl = mptn->mctl;
  REP_MCTL_CHECK (rmctl);

  /*
    Before starting checks, create a rep_ictl.
  */
  rictl = g_new0 (rep_ictl_t, 1);
  rictl->magic = REP_ICTL_MAGIC;
  rictl->mctl = rmctl;
  rictl->begin = begin;
  rictl->end = end;
  rictl->vars = vars;
  rictl->stage = stage;
  rictl->data = data;
  rictl->bb = NULL;
  rictl->ectl = NULL;
  rictl->sub_ictl = NULL;

  /* 
     First, check if the string has zero length and if this is allowed.
     Second, check if there may be just one occurrence of Sub.
     Third, start a loop for splitting the string into parts.
      (the second and the third steps are actually done inside 
      rep_matcher_iter_step).
  */

  if (end == begin && rmctl->zero_allowed)
    {
      /* return a half-baked rictl */
      return rictl;
    }
  else
    {
      /* 
	 Make the first step of the iterator.  Given an unfilled rictl, it
	 will fill it.
      */
      return rep_matcher_iter_step (rictl);
    }
}

/*
  The step routine does not have to take zero_allowed into account
*/
static gpointer 
rep_matcher_iter_step (gpointer mictl)
{
  rep_ictl_t *rictl = mictl;
  rep_mctl_t *rmctl;
  chr *begin;
  chr *end;
  mptn_t *sub;

  g_assert (mictl != NULL);
  REP_ICTL_CHECK (rictl);

  rmctl = rictl->mctl;
  begin = rictl->begin;
  end = rictl->end;
  sub = rmctl->sub;

  /*
    At the beginning, we need to check whether the subexpression
    (repeated just once) matches the string.
  */
  if (!rictl->sub_ictl)
    /* Have not started working with occurrences after the first */
    {
      if (!rictl->ectl)
	/* First invocation */
	{
	  if (sub->min_len <= end - begin 	/* The length is approporiate */
	      && sub->max_len >= end - begin
	      && (rictl->ectl 
		  = mptn_exec_start_with_vars (sub, begin, end,
					       rictl->vars,
					       rictl->stage + 1, /* Not sure whether we need to increase Stage -- Goga */
					       rictl->data)))
	    return rictl; 	/* Alright, it matches */
	}
      else
	/* Continuing */
	{
	  if ((rictl->ectl = mptn_exec_step (rictl->ectl)))
	    return rictl; 	/* Matches continue */
	}

      /* 
	 Here, the possibilities to match the whole string against Sub
	 are exhausted.  We need to start checking for real repetition.
      
	 Rictl->bb is set one chr to the left from the place where we
	 should really start looking (to fall gracefully into the For loop
	 below).
      */
      rictl->bb = begin + MAX (0, sub->min_len - 1);
    }

  /* 
     Now we try to work with multiple repetitions
     (only if this is allowed).

     First, we may just need to advance the subordinated 
     ictl (for all the occurrences except the first)
  */
  if (rmctl->rep_allowed)
    {
      if (rictl->sub_ictl 
	  && (rictl->sub_ictl = rep_matcher_iter_step (rictl->sub_ictl)))
	return rictl;
      
      /* 
	 Second, we may need to advance rictl->ectl
	 (for the first occurrence).
	 In this case we need to restart Sub_ictl.
      */
      else if (rictl->ectl
	       && (rictl->ectl = mptn_exec_step (rictl->ectl))
	       && (rictl->sub_ictl = rep_sub_ictl_start (rictl)))
	return rictl;

      /*
	We need to advance the boundary between the first occurrence 
	and the rest.
      */
      else
	{
	  for (rictl->bb++;
	       rictl->bb - begin <= sub->max_len
		 && end - rictl->bb >= sub->min_len
		 && end - rictl->bb > 0 	
		 /* This special case is needed for the case when
		    sub->min_len == 0 */
		 ;
	       rictl->bb++)
	    {
	      /* ectl may be left from previous attempts */
	      if (rictl->ectl) mptn_ectl_free (rictl->ectl);
	      
	      if ((rictl->ectl = mptn_exec_start_with_vars (sub, begin, rictl->bb,
							    rictl->vars,
							    rictl->stage+1,
							    rictl->data))
		  && (rictl->sub_ictl = rep_sub_ictl_start (rictl)))
		return rictl; 	/* Found a new match */
	    }
	}
    }

  /* No matches left */
  rep_matcher_ictl_free (rictl);
  return NULL;
}

/*
  Create a Rictl for the matches after the first one.
  Rictl should have the Ectl and Bb fields filled.
*/
static rep_ictl_t *
rep_sub_ictl_start (rep_ictl_t *rictl)
{
  rep_ictl_t *res;

  REP_ICTL_CHECK (rictl);
  g_assert (rictl->ectl != NULL);
  g_assert (rictl->bb != NULL);

  /* Allocate the structure and fill in the fields */
  res = g_new0 (rep_ictl_t, 1);
  res->magic = REP_ICTL_MAGIC;
  res->mctl = rictl->mctl;
  res->begin = rictl->bb;
  res->end = rictl->end;
  res->vars = rictl->vars;
  res->stage = mptn_ectl_stage (rictl->ectl);
  res->data = rictl->data;
  res->bb = NULL;
  res->ectl = NULL;
  res->sub_ictl = NULL;

  /* This is the real start of the subordinated iterator */
  return rep_matcher_iter_step (res);
}

static void 
rep_matcher_ictl_free (gpointer mictl)
{
  rep_ictl_t *rictl = mictl;

  g_assert (rictl != NULL);
  REP_ICTL_CHECK (rictl);

  if (rictl->ectl)
    mptn_ectl_free (rictl->ectl);

  if (rictl->sub_ictl)
    rep_matcher_ictl_free (rictl->sub_ictl);

  g_free (rictl);
}

static guint 
rep_matcher_stage (gpointer mictl)
{
  rep_ictl_t *rictl = mictl;

  g_assert (rictl != NULL);
  REP_ICTL_CHECK (rictl);

  /* Check if a subordinated Rep_ictl is running */
  if (rictl->sub_ictl)
    return rep_matcher_stage (rictl->sub_ictl);

  /* Check if a subordinated Ectl is running */
  if (rictl->ectl)
    return mptn_ectl_stage (rictl->ectl);

  /* 
     (Must be the stage for allowed zero)
     Just return the stage we were started at.
  */
  return rictl->stage;
}

/*
  The subst function only attempts to work when there is zero-or-one "repetition".
  In this case it attempts to substitute the subexpression; if it fails, it returns
  an empty string.
*/
static chr *
rep_matcher_subst (mptn_t *mptn, mptn_var_t *vars, gpointer data)
{
  rep_mctl_t *rmctl;
  chr *res;

  MPTN_CHECK (mptn);

  rmctl = (rep_mctl_t*) mptn->mctl;
  REP_MCTL_CHECK (rmctl);

  if (rmctl->rep_allowed)
    return NULL; 		/* impossible to recover */

  if ((res = mptn_subst (rmctl->sub, vars, data)))
    return res;
  else if (rmctl->zero_allowed)
    return  g_new0 (chr, 1);	/* A way to get a zero length string */
  else	/* z_a == FALSE, r_a == FALSE makes no sense, but why not make it work? */
    return NULL;
}

static void 
rep_matcher_print (mptn_t *mptn, FILE *fp, guint offset)
{
  rep_mctl_t *rmctl;

  MPTN_CHECK (mptn);
  g_assert (fp != NULL);

  rmctl = mptn->mctl;
  REP_MCTL_CHECK (rmctl);

  make_offset (fp, offset);
  fprintf (fp, "Rep(%s, %s)\n",
	   rmctl->zero_allowed ? "Zero allowed" : "Zero not allowed",
	   rmctl->rep_allowed ? "Rep allowed" : "Rep not allowed");

  mptn_print (rmctl->sub, fp, offset+2);
}

/**************************** Variable ****************************************/
static void var_matcher_destroy (gpointer mctl);
static gpointer var_matcher_iter_start (mptn_t *mptn,
					chr *begin, chr *end,
					mptn_var_t *vars, guint stage,
					gpointer data);
static chr *var_matcher_subst (mptn_t *mptn, mptn_var_t *vars, gpointer data);
static void var_matcher_print (mptn_t *mptn, FILE *fp, guint offset);

mptn_matcher_ops_t mptn_var_matcher_ops = {
  NULL,
  var_matcher_destroy,
  var_matcher_iter_start,
  NULL,				/* iter_step -- just one match */
  NULL,				/* ictl_free */
  NULL,				/* matcher_stage */
  var_matcher_subst,
  var_matcher_print
};

/* Structure corresponding to Mctl for Var_matcher */
#define VAR_MCTL_MAGIC 0x01159344

typedef struct var_mctl {
  guint magic;
  int ix;		  /* Index inside the variables array */
  mptn_t *vmptn;	  /* Restriction for variable value */
  GQuark vcode;	  	  /* Representing the variable's name */
} var_mctl_t;

#define VAR_MCTL_CHECK(ic)			\
	do {					\
	  if ((ic)->magic != VAR_MCTL_MAGIC)	\
	    g_error ("Not a var_mctl");	\
	} while(0)

/*
  Initialization
*/
mptn_t *
mptn_var_matcher_init (mptn_control_t *control, 
		       chr *args_begin, chr *args_end,
		       GArray *varnames,
		       gpointer data)
{
  mptn_t *res;
  guint ix;
  GQuark vcode;
  chr *vname;
  mptn_t *vmptn;
  var_mctl_t *vmctl;

  g_assert (args_begin != NULL);
  control = mptn_control_normalize (control);

  if (!args_end)
    args_end = args_begin + chra_len (args_begin);

  vname = chra_ndup (args_begin, args_end - args_begin);
  vcode = chra_quark_from_string (vname);
  g_free (vname);

  res = mptn_new (control);
  res->ops = &mptn_var_matcher_ops;

  ix = mptn_var_get_index (varnames, vcode);
  vmptn = mptn_control_get_var_ptn (control, vcode);

  vmctl = g_new (var_mctl_t, 1);
  vmctl->magic = VAR_MCTL_MAGIC;
  vmctl->ix = ix;
  vmctl->vmptn = vmptn;
  vmctl->vcode = vcode;

  res->mctl = vmctl;
  
  res->min_len = vmptn ? vmptn->min_len : 0;
  res->max_len = vmptn ? vmptn->max_len : G_MAXINT;

  if (vmptn)
    mptn_refinc (vmptn);

  return res;
}

static void 
var_matcher_destroy (gpointer mctl)
{
  var_mctl_t *vmctl = mctl;

  VAR_MCTL_CHECK (vmctl);

  if (vmctl->vmptn)
    mptn_free (vmctl->vmptn);

  g_free (vmctl);
}

static gpointer 
var_matcher_iter_start (mptn_t *mptn,
			chr *begin, chr *end,
			mptn_var_t *vars, guint stage,
			gpointer data)
{
  var_mctl_t *vmctl;
  mptn_var_t *var;

  MPTN_CHECK (mptn);

  vmctl = mptn->mctl;
  VAR_MCTL_CHECK (vmctl);

  g_assert (begin);
  g_assert (end);
  g_assert (vars);

  var = vars + vmctl->ix;

  if (var->beg)
    {
      g_assert (var->vcode == vmctl->vcode);

      if (var->end - var->beg == end - begin
	  && chra_ncmp (var->beg, begin, end - begin) == 0)
	return (gpointer)1;
      else
	return NULL;
    }
  else
    {
      g_assert (var->vcode == vmctl->vcode);

      if (vmctl->vmptn		/* Check the restriction */
	  && !mptn_exec_one (vmctl->vmptn, begin, end,	/* mptn, begin, end */
			     0, NULL, 			/* nvars, vars */
			     NULL))			/* data */
	return NULL;

      var->beg = begin;
      var->end = end;
      var->stage = stage;
      var->allocated = FALSE;

      return (gpointer) 1;
    }
}
	
static chr *
var_matcher_subst (mptn_t *mptn, mptn_var_t *vars, gpointer data)
{
  var_mctl_t *vmctl;
  mptn_var_t *vv;

  MPTN_CHECK (mptn);

  vmctl = mptn->mctl;
  VAR_MCTL_CHECK (vmctl);

  if (!vars)
    return NULL;	/* no vars, no substitution */

  /* Linear search :( */
  for (vv = vars; vv->vcode; vv++)
    if (vv->vcode == vmctl->vcode) 	/* found */
      {
	if (!vv->beg)
	  return NULL;		/* No value for the variable */
	else
	  return chra_ndup (vv->beg, vv->end - vv->beg);
      }

  return NULL;
}

static void
var_matcher_print (mptn_t *mptn, FILE *fp, guint offset)
{
  var_mctl_t *vmctl;

  MPTN_CHECK (mptn);

  vmctl = mptn->mctl;
  VAR_MCTL_CHECK (vmctl);

  make_offset (fp, offset);
  fprintf (fp, "Var(%s)\n", chra_quark_to_string(vmctl->vcode));
}

/*********** Matcher making all the internal assignments temporary ************/
static void ta_matcher_destroy (gpointer mctl);
static gpointer ta_matcher_iter_start (mptn_t *mptn,
				       chr *begin, chr *end,
				       mptn_var_t *vars, guint stage,
				       gpointer data);
static chr *ta_matcher_subst (mptn_t *mptn, mptn_var_t *vars, gpointer data);
static void ta_matcher_print (mptn_t *mptn, FILE *fp, guint offset);

mptn_matcher_ops_t mptn_ta_matcher_ops = {
  NULL,
  ta_matcher_destroy,
  ta_matcher_iter_start,
  NULL,				/* iter_step -- just one match */
  NULL,				/* ictl_free */
  NULL,				/* matcher_stage */
  NULL, 			/* not recoverable */
  ta_matcher_print
};

/*
  Initialization
*/
mptn_t *
mptn_ta_matcher_init (mptn_control_t *control, 
		      mptn_t *sub)
{
  mptn_t *res;

  res = mptn_new (control);
  res->ops = &mptn_ta_matcher_ops;
  res->min_len = sub->min_len;
  res->max_len = sub->max_len;
  res->flags = 0;
  res->mctl = sub;

  return res;
}

static void 
ta_matcher_destroy (gpointer mctl)
{
  mptn_free ((mptn_t*)mctl);
}

static gpointer 
ta_matcher_iter_start (mptn_t *mptn,
		       chr *begin, chr *end,
		       mptn_var_t *vars, guint stage,
		       gpointer data)
{
  mptn_ectl_t *ectl;

  MPTN_CHECK (mptn);
  MPTN_CHECK ((mptn_t*)mptn->mctl);

  /* First, start the iterator for the subexpression */
  ectl = mptn_exec_start_with_vars ((mptn_t*)mptn->mctl,
				    begin, end,
				    vars, stage,
				    data);
  
  if (!ectl)
    return NULL;
  else
    {
      /* Forget the results */
      mptn_vars_clear (vars, stage);
      mptn_ectl_free (ectl);
      return (gpointer)1;
    }
}

static void
ta_matcher_print (mptn_t *mptn, FILE *fp, guint offset)
{
  MPTN_CHECK (mptn);
  MPTN_CHECK ((mptn_t*)mptn->mctl);

  make_offset (fp, offset);
  fprintf (fp, "Temporary_assignment\n");
  mptn_print ((mptn_t*)mptn->mctl, fp, offset+2);
}


/* Auxiliary function */
static void
make_offset (FILE *fp, guint offset)
{
  int i;
  
  for (i = 0; i < offset; i++)
    fputc (' ', fp);
}
