/* Mptn -- regexp-like pattern matcher suited for morphological analysis
 * Copyright (C) 2000  George Bronnikov
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
  Private header.

  $Id: mptn-private.h 1.3 Sun, 23 Jan 2000 20:56:51 -0600 goga $
*/
#ifndef MPTN_PRIVATE_H
#define MPTN_PRIVATE_H

#ifndef MPTN_H
#  error "Mptn-private.h should be included after mptn.h"
#endif
#ifndef MPTN_CHRSET_H
#  error "Mptn-private.h should be included after mptn-chrset.h"
#endif

/* Mptn_control_t */
#define MPTN_CONTROL_MAGIC 0x212850a

struct mptn_control {
  guint magic;
  GHashTable *vars_htable;
  GHashTable *matchers_htable;  /* Should we have separate namespaces for them? */
  GHashTable *matcher_param_htable;
};

#define MPTN_CONTROL_CHECK(c)				\
	do {						\
	  if ((c)->magic != MPTN_CONTROL_MAGIC)		\
	    g_error ("Not an mptn_control_t");		\
        } while (0)

/* Make sure control refers to something -- called at entry to toplevel functons */
mptn_control_t *mptn_control_normalize (mptn_control_t *control);

/* Get the mptn corresponding to a variable */
mptn_t *mptn_control_get_var_ptn (mptn_control_t *control, GQuark vcode);

/* Get the matcher function corresponding to a matcher's name */
mptn_matcher_ops_t *mptn_control_get_matcher (mptn_control_t *control, GQuark mcode);

/* Get the matcher parameter */
gpointer mptn_control_get_matcher_param (mptn_control_t *control, GQuark mcode);

/* Clear the values of the variables younger than Stage */
void mptn_vars_clear (mptn_var_t *vars, guint stage);

/* Mptn_t */
#define MPTN_MAGIC	0x212850b

#define MPTN_CHECK(m)				\
        do {					\
	  if ((m)->magic != MPTN_MAGIC)		\
	    g_error ("Not an Mptn_t");		\
        } while (0)
/*
  Allocation and basic setup.
*/
mptn_t * mptn_new (mptn_control_t *control);

/* Mptn_ectl_t */
#define MPTN_ECTL_MAGIC 0x212850c

struct mptn_ectl {
  guint magic;
  guint flags;
#define MPTN_ECTL_VARS_ALLOCATED 	0x01 /* Need to free when the loop is done? */

  mptn_t *mptn;
  mptn_var_t *vars;
  guint    stage;		/* Stage of the matching */
  gpointer ictl;		/* What is passed to the actual matcher */
};

#define MPTN_ECTL_CHECK(e)			\
	do {					\
	  if ((e)->magic != MPTN_ECTL_MAGIC)	\
	    g_error ("Not an Mptn_ectl_t");	\
	} while (0)

/*
  Predefined matchers.
*/

/* Constant string */
mptn_t *mptn_string_matcher_init (mptn_control_t *control, 
				  chr *args_begin, chr *args_end,
				  GArray *varnames,
				  gpointer data);
extern mptn_matcher_ops_t mptn_string_matcher_ops;

/* Anything goes */
mptn_t *mptn_any_matcher_init (mptn_control_t *control,
			       gboolean empty_allowed, gboolean repetition);
extern mptn_matcher_ops_t mptn_any_matcher_ops;

/* Alternative */
mptn_t * mptn_alt_matcher_init (mptn_control_t *control,  mptn_t **alts);
extern mptn_matcher_ops_t mptn_alt_matcher_ops;

/* Conjunction */
mptn_t * mptn_conj_matcher_init (mptn_control_t *control,  mptn_t **subs);
extern mptn_matcher_ops_t mptn_conj_matcher_ops;

/* Concatenation */
mptn_t * mptn_cat_matcher_init (mptn_control_t *control,  mptn_t **subs);
extern mptn_matcher_ops_t mptn_cat_matcher_ops;

/* Charset */
mptn_t * mptn_mptn_chrset_init (mptn_control_t *control, mptn_chrset_t *chrset);
extern mptn_matcher_ops_t mptn_mptn_chrset_matcher_ops;

/* Repetition */
mptn_t *mptn_rep_matcher_init (mptn_control_t *control, mptn_t *sub, 
			       gboolean zero_allowed, gboolean rep_allowed);
extern mptn_matcher_ops_t mptn_rep_matcher_ops;

/* Variable */
mptn_t *mptn_var_matcher_init (mptn_control_t *control, 
			       chr *args_begin, chr *args_end,
			       GArray *varnames,
			       gpointer data);
extern mptn_matcher_ops_t mptn_var_matcher_ops;

/* Temporary var assignment */
mptn_t *mptn_ta_matcher_init (mptn_control_t *control, mptn_t *sub);
extern mptn_matcher_ops_t mptn_ta_matcher_ops;

/******* Aux functions *********/

/*
  Find a character contained in What, outside the brackets,
  where Leftb serve as left brackets, Rightb as right brackets,
  Esc serving as escape characters.
  (Exact balance of brackets is not counted; thus, [) would
  count as valid pair.

  Return a pointer to the character found, or NULL on failure.
*/
chr *maux_find (chr *str, chr *what, chr *leftb, chr *rightb, chr *esc);

#endif /* already included */
